@@ -8,7 +8,6 @@ import shlex
 from subprocess import PIPE, run
 from dataclasses import dataclass, asdict
 from typing import Any, Optional
-from packaging.version import parse as parse_version
 
 PROGNAME = "trurl"
 TESTFILE = "tests.json"
@@ -26,6 +25,17 @@ class CommandOutput:
     stderr: str
 
 
+def testComponent(value, exp):
+    if isinstance(exp, bool):
+        result = value == 0 or value not in ("", [])
+        if exp:
+            return result
+        else:
+            return not result
+
+    return value == exp
+
+
 class TestCase:
     def __init__(self, testIndex, baseCmd, **testCase):
         self.testIndex = testIndex
@@ -52,14 +62,14 @@ class TestCase:
             encoding="utf-8"
         )
 
-        if type(self.expected["stdout"]) == str:
-            stdout = output.stdout
-        else:
+        if isinstance(self.expected["stdout"], list):
             # if we dont expect string, parse to json
             try:
                 stdout = json.loads(output.stdout)
             except json.decoder.JSONDecodeError:
                 stdout = None
+        else:
+            stdout = output.stdout
 
         # assume stderr is always going to be string
         stderr = output.stderr
@@ -69,13 +79,9 @@ class TestCase:
 
     def test(self):
         # return true only if stdout, stderr and errorcode is equal to the ones found in testfile
-
-        tests = []
-        for item in self.expected:
-            passed = self.expected[item] == asdict(self.commandOutput)[item]
-            tests.append(passed)
-
-        self.testPassed = all(tests)
+        self.testPassed = all(
+            testComponent(asdict(self.commandOutput)[k], exp)
+            for k, exp in self.expected.items())
         return self.testPassed
 
     def printVerbose(self):
@@ -83,17 +89,20 @@ class TestCase:
         self.printConcise()
         print(NOCOLOR, end="")
 
-        for item in self.expected:
-            itemFail = self.expected[item] != asdict(self.commandOutput)[item]\
-                    and self.commandOutput.returncode == 1
+        for component, exp in self.expected.items():
+            value = asdict(self.commandOutput)[component]
+            itemFail = self.commandOutput.returncode == 1 and \
+                not testComponent(value, exp)
 
-            print(f"--- {item} --- ")
+            print(f"--- {component} --- ")
             print("expected:")
-            print(f"{self.expected[item]!r}")
+            print("nothing" if exp is False else
+                  "something" if exp is True else
+                  f"{exp!r}")
             print("got:")
             if itemFail:
                 print(RED, end="")
-            print(f"{asdict(self.commandOutput)[item]!r}")
+            print(f"{value!r}")
             if itemFail:
                 print(NOCOLOR, end="")
 
@@ -124,11 +133,7 @@ def main(argc, argv):
             stdout=PIPE, stderr=PIPE,
             encoding="utf-8"
         )
-        version_re = r' libcurl\/([\w\.\-]+) \[built\-with ([\w\.\-]+)\]$'
-        match = re.findall(version_re, output.stdout)
-        libcurl_runtime, libcurl_buildtime = match[0]
-        libcurl_runtime = parse_version(libcurl_runtime)
-        libcurl_buildtime = parse_version(libcurl_buildtime)
+        features = output.stdout.split('\n')[1].split()[1:]
 
         with open(path.join(baseDir, TESTFILE), "r") as file:
             allTests = json.load(file)
@@ -156,24 +161,11 @@ def main(argc, argv):
         numTestsPassed = 0
         numTestsSkipped = 0
         for testIndex in testIndexesToRun:
-            # skip tests if the run-time libcurl version is too low
-            if "minruntime" in allTests[testIndex]:
-                minruntime = allTests[testIndex]["minruntime"]
-                minruntime = parse_version(minruntime)
-                if libcurl_runtime < minruntime:
-                    skipMessage = "libcurl run-time version is too low: "
-                    skipMessage += f"{libcurl_runtime} < {minruntime}"
-                    print(f"{testIndex + 1}: skipped\t{skipMessage}")
-                    numTestsSkipped += 1
-                    continue
-            # skip tests if the build-time libcurl version is too low
-            if "minbuildtime" in allTests[testIndex]:
-                minbuildtime = allTests[testIndex]["minbuildtime"]
-                minbuildtime = parse_version(minbuildtime)
-                if libcurl_buildtime < minbuildtime:
-                    skipMessage = "libcurl build-time version is too low: "
-                    skipMessage += f"{libcurl_buildtime} < {minbuildtime}"
-                    print(f"{testIndex + 1}: skipped\t{skipMessage}")
+            # skip tests if required features are not met
+            if "required" in allTests[testIndex]:
+                required = allTests[testIndex]["required"]
+                if not set(required).issubset(set(features)):
+                    print(f"Missing feature, skipping test {testIndex + 1}.")
                     numTestsSkipped += 1
                     continue
 
