@@ -22,6 +22,7 @@
  *
  ***************************************************************************/
 
+#include <errno.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -39,6 +40,9 @@
 #define strdup _strdup
 #endif
 
+#if CURL_AT_LEAST_VERSION(7,77,0)
+#define SUPPORTS_NORM_IPV4
+#endif
 #if CURL_AT_LEAST_VERSION(7,81,0)
 #define SUPPORTS_ZONEID
 #endif
@@ -72,13 +76,18 @@
 
 #define REPLACE_NULL_BYTE '.' /* for query:key extractions */
 
+enum {
+  VARMODIFIER_URLDECODED = 1 << 1,
+  VARMODIFIER_DEFAULT    = 1 << 2,
+  VARMODIFIER_PUNY       = 1 << 3,
+};
+
 struct var {
   const char *name;
   CURLUPart part;
 };
 
 static const struct var variables[] = {
-  {"url",      CURLUPART_URL},
   {"scheme",   CURLUPART_SCHEME},
   {"user",     CURLUPART_USER},
   {"password", CURLUPART_PASSWORD},
@@ -141,42 +150,47 @@ static void errorf(int exit_code, char *fmt, ...)
 
 #define VERIFY(op, exit_code, ...) \
   do { \
-    if(op->verify) { \
+    if(!op->verify) \
+      warnf(__VA_ARGS__); \
+    else { \
       /* make sure to terminate the JSON array */ \
       if(op->jsonout) \
-        fputs("\n]\n", stdout); \
+        printf("%s]\n", op->urls ? "\n" : ""); \
       errorf(exit_code, __VA_ARGS__); \
-    } else \
-      warnf(__VA_ARGS__); \
+    } \
   } while(0)
 
 static void help(void)
 {
   int i;
-  fprintf(stderr, "Usage: " PROGNAME " [options] [URL]\n"
-          "  -a, --append [component]=[data] - append data to component\n"
-          "  --accept-space               - give in to this URL abuse\n"
-          "  -f, --url-file [file/-]      - read URLs from file or stdin\n"
-          "  -g, --get [{component}s]     - output component(s)\n"
-          "  -h, --help                   - this help\n"
-          "  --iterate [component]=[list] - create multiple URL outputs\n"
-          "  --json                       - output URL as JSON\n"
-          "  --query-separator [letter]   - if something else than '&'\n"
-          "  --redirect [URL]             - redirect to this\n"
-          "  -s, --set [component]=[data] - set component content\n"
-          "  --sort-query                 - alpha-sort the query pairs\n"
-          "  --trim [component]=[what]    - trim component\n"
-          "  --url [URL]                  - URL to work with\n"
-          "  -v, --version                - show version\n"
-          "  --verify                     - return error on (first) bad URL\n"
-          " URL COMPONENTS:\n"
-          "  "
-    );
+  fputs(
+    "Usage: " PROGNAME " [options] [URL]\n"
+    "  -a, --append [component]=[data]  - append data to component\n"
+    "      --accept-space               - give in to this URL abuse\n"
+    "      --default-port               - add known default ports\n"
+    "  -f, --url-file [file/-]          - read URLs from file or stdin\n"
+    "  -g, --get [{component}s]         - output component(s)\n"
+    "  -h, --help                       - this help\n"
+    "      --iterate [component]=[list] - create multiple URL outputs\n"
+    "      --json                       - output URL as JSON\n"
+    "      --keep-port                  - keep known default ports\n"
+    "      --no-guess-scheme            - require scheme in URLs\n"
+    "      --punycode                   - encode hostnames in punycode\n"
+    "      --query-separator [letter]   - if something else than '&'\n"
+    "      --redirect [URL]             - redirect to this\n"
+    "  -s, --set [component]=[data]     - set component content\n"
+    "      --sort-query                 - alpha-sort the query pairs\n"
+    "      --trim [component]=[what]    - trim component\n"
+    "      --url [URL]                  - URL to work with\n"
+    "  -v, --version                    - show version\n"
+    "      --verify                     - return error on (first) bad URL\n"
+    " URL COMPONENTS:\n"
+    "  ", stdout);
   for(i = 0; i< NUM_COMPONENTS; i++) {
-    fprintf(stderr, "%s%s", i?", ":"", variables[i].name);
+    printf("%s%s", i?", ":"", variables[i].name);
   }
-  fprintf(stderr, "\n");
-  exit(1);
+  fputs("\n", stdout);
+  exit(0);
 }
 
 static void show_version(void)
@@ -184,6 +198,32 @@ static void show_version(void)
   curl_version_info_data *data = curl_version_info(CURLVERSION_NOW);
   fprintf(stdout, "%s version %s libcurl/%s [built-with %s]\n",
           PROGNAME, TRURL_VERSION_TXT, data->version, LIBCURL_VERSION);
+  /* puny code isn't guaranteed based on the version, so it must be polled
+   * from libcurl */
+  bool supports_puny = false;
+#ifdef SUPPORTS_PUNYCODE
+  const char *const *feature_name = data->feature_names;
+  while(*feature_name && !supports_puny) {
+    supports_puny = !strncmp(*feature_name, "IDN", 3);
+    feature_name++;
+  }
+#endif
+
+  fprintf(stdout, "features: %s", supports_puny?"punycode ":"");
+#ifdef SUPPORTS_ALLOW_SPACE
+  fprintf(stdout, "white-space ");
+#endif
+#ifdef SUPPORTS_ZONEID
+  fprintf(stdout, "zone-id ");
+#endif
+#ifdef SUPPORTS_URL_STRERROR
+  fprintf(stdout, "url-strerror ");
+#endif
+#ifdef SUPPORTS_NORM_IPV4
+  fprintf(stdout, "normalize-ipv4");
+#endif
+
+  fprintf(stdout, "\n");
   exit(0);
 }
 
@@ -210,7 +250,11 @@ struct option {
   bool jsonout;
   bool verify;
   bool accept_space;
+  bool default_port;
+  bool keep_port;
+  bool punycode;
   bool sort_query;
+  bool no_guess_scheme;
   bool end_of_options;
   unsigned char output;
 
@@ -357,32 +401,32 @@ static int getarg(struct option *op,
     help();
   else if(checkoptarg("--url", flag, arg)) {
     urladd(op, arg);
-    *usedarg = 1;
+    *usedarg = true;
   }
   else if(checkoptarg("-f", flag, arg) ||
           checkoptarg("--url-file", flag, arg)) {
     urlfile(op, arg);
-    *usedarg = 1;
+    *usedarg = true;
   }
   else if(checkoptarg("-a", flag, arg) ||
           checkoptarg("--append", flag, arg)) {
     appendadd(op, arg);
-    *usedarg = 1;
+    *usedarg = true;
   }
   else if(checkoptarg("-s", flag, arg) ||
           checkoptarg("--set", flag, arg)) {
     setadd(op, arg);
-    *usedarg = 1;
+    *usedarg = true;
   }
   else if(checkoptarg("--iterate", flag, arg)) {
     iteradd(op, arg);
-    *usedarg = 1;
+    *usedarg = true;
   }
   else if(checkoptarg("--redirect", flag, arg)) {
     if(op->redirect)
       errorf(ERROR_FLAG, "only one --redirect is supported");
     op->redirect = arg;
-    *usedarg = 1;
+    *usedarg = true;
   }
   else if(checkoptarg("--query-separator", flag, arg)) {
     if(op->qsep)
@@ -390,11 +434,11 @@ static int getarg(struct option *op,
     if(strlen(arg) != 1)
       errorf(ERROR_FLAG, "only single-letter query separators are supported");
     op->qsep = arg;
-    *usedarg = 1;
+    *usedarg = true;
   }
   else if(checkoptarg("--trim", flag, arg)) {
     trimadd(op, arg);
-    *usedarg = 1;
+    *usedarg = true;
   }
   else if(checkoptarg("-g", flag, arg) ||
           checkoptarg("--get", flag, arg)) {
@@ -403,7 +447,7 @@ static int getarg(struct option *op,
     if(op->jsonout)
       errorf(ERROR_FLAG, "--get is mututally exclusive with --json");
     op->format = arg;
-    *usedarg = 1;
+    *usedarg = true;
   }
   else if(!strcmp("--json", flag)) {
     if(op->format)
@@ -419,6 +463,14 @@ static int getarg(struct option *op,
     warnf("built with too old libcurl version, --accept-space does not work");
 #endif
   }
+  else if(!strcmp("--default-port", flag))
+    op->default_port = true;
+  else if(!strcmp("--keep-port", flag))
+    op->keep_port = true;
+  else if(!strcmp("--punycode", flag))
+    op->punycode = true;
+  else if(!strcmp("--no-guess-scheme", flag))
+    op->no_guess_scheme = true;
   else if(!strcmp("--sort-query", flag))
     op->sort_query = true;
   else
@@ -426,8 +478,8 @@ static int getarg(struct option *op,
   return 0;
 }
 
-static void showqkey(const char *key, size_t klen, bool urldecode,
-                     bool showall)
+static void showqkey(FILE *stream, const char *key, size_t klen,
+                     bool urldecode, bool showall)
 {
   int i;
   bool shown = false;
@@ -437,8 +489,8 @@ static void showqkey(const char *key, size_t klen, bool urldecode,
     if(!strncmp(key, qp[i], klen) &&
        (qp[i][klen] == '=')) {
       if(shown)
-        fputc(' ', stdout);
-      fputs(&qp[i][klen + 1], stdout);
+        fputc(' ', stream);
+      fputs(&qp[i][klen + 1], stream);
       if(!showall)
         break;
       shown = true;
@@ -457,6 +509,37 @@ static const struct var *comp2var(const char *name, size_t vlen)
   return NULL;
 }
 
+static CURLUcode geturlpart(struct option *o, int modifiers, CURLU *uh,
+                            CURLUPart part, char **out)
+{
+  return curl_url_get(uh, part, out,
+                      (((modifiers & VARMODIFIER_DEFAULT) ||
+                        o->default_port) ?
+                       CURLU_DEFAULT_PORT :
+                       ((part != CURLUPART_URL || o->keep_port) ?
+                        0 : CURLU_NO_DEFAULT_PORT))|
+#ifdef SUPPORTS_PUNYCODE
+                      (((modifiers & VARMODIFIER_PUNY) || o->punycode) ?
+                       CURLU_PUNYCODE : 0)|
+#endif
+                      CURLU_NON_SUPPORT_SCHEME|
+                      ((modifiers & VARMODIFIER_URLDECODED) ?
+                       CURLU_URLDECODE : 0));
+}
+
+static void showurl(FILE *stream, struct option *o, int modifiers,
+                    CURLU *uh)
+{
+  char *url;
+  CURLUcode rc = geturlpart(o, modifiers, uh, CURLUPART_URL, &url);
+  if(rc) {
+    VERIFY(o, ERROR_BADURL, "invalid url [%s]", curl_url_strerror(rc));
+    return;
+  }
+  fputs(url, stream);
+  curl_free(url);
+}
+
 static void get(struct option *op, CURLU *uh)
 {
   FILE *stream = stdout;
@@ -481,12 +564,12 @@ static void get(struct option *op, CURLU *uh)
       }
       else {
         /* this is meant as a variable to output */
+        const char *start = ptr;
         char *end;
         char *cl;
         size_t vlen;
-        bool urldecode = true;
-        bool punycode = false;
-        bool handled = true;
+        bool isquery = false;
+        int mods = 0;
         end = strchr(ptr, endbyte);
         ptr++; /* pass the { */
         if(!end) {
@@ -494,49 +577,61 @@ static void get(struct option *op, CURLU *uh)
           fputc(startbyte, stream);
           continue;
         }
+
         /* {path} {:path} */
         if(*ptr == ':') {
-          urldecode = false;
+          mods |= VARMODIFIER_URLDECODED;
           ptr++;
         }
         vlen = end - ptr;
-        /* check for the last colon within here */
-        cl = memchr(ptr, ':', vlen);
-        if(cl) {
-          /* deduct the colon part */
-          if(!strncmp(ptr, "query-all:", 10))
-            showqkey(&ptr[10], end - cl - 1, urldecode, true);
-          else if(!strncmp(ptr, "query:", 6))
-            showqkey(&ptr[6], end - cl - 1, urldecode, false);
-          else if(!strncmp(ptr, "puny:", 5)) {
-            punycode = true;
-#ifndef SUPPORTS_PUNYCODE
-            warnf("Built without punycode support");
-#endif
-            ptr = cl + 1;
-            vlen = end - ptr;
-            handled = false;
+        do {
+          cl = memchr(ptr, ':', vlen);
+          if(!cl)
+            break;
+
+          /* modifiers! */
+          if(!strncmp(ptr, "default:", cl - ptr + 1))
+            mods |= VARMODIFIER_DEFAULT;
+          else if(!strncmp(ptr, "puny:", cl - ptr + 1))
+            mods |= VARMODIFIER_PUNY;
+          else {
+            /* {query: or {query-all: */
+            if(!strncmp(ptr, "query-all:", cl - ptr + 1)) {
+              showqkey(stream, cl + 1, end - cl - 1,
+                       (mods & VARMODIFIER_URLDECODED) == 0, true);
+            }
+            else if(!strncmp(ptr, "query:", cl - ptr + 1)) {
+              isquery = true;
+              showqkey(stream, cl + 1, end - cl - 1,
+                       (mods & VARMODIFIER_URLDECODED) == 0, false);
+            }
+            else {
+              /* syntax error */
+              vlen = 0;
+              end[1] = '\0';
+              break;
+            }
+            isquery = true;
           }
-          else
-            errorf(ERROR_GET, "Bad --get syntax: %s", ptr);
-        }
-        else
-          handled = false;
-        if(!handled) {
+
+          ptr = cl + 1;
+          vlen = end - ptr;
+        } while(true);
+
+        if(isquery)
+          ;
+        else if(!vlen)
+          errorf(ERROR_GET, "Bad --get syntax: %s", start);
+        else if(!strncmp(ptr, "url", vlen))
+          showurl(stream, op, mods, uh);
+        else {
           const struct var *v = comp2var(ptr, vlen);
           if(v) {
             char *nurl;
-            CURLUcode rc;
-            rc = curl_url_get(uh, v->part, &nurl,
-                              CURLU_DEFAULT_PORT|
-                              CURLU_NO_DEFAULT_PORT|
-#ifdef SUPPORTS_PUNYCODE
-                              (punycode?CURLU_PUNYCODE:0)|
-#endif
-                              (urldecode?CURLU_URLDECODE:0));
+            CURLUcode rc = geturlpart(op, mods, uh, v->part, &nurl);
             switch(rc) {
             case CURLUE_OK:
-              fprintf(stream, "%s", nurl);
+              fputs(nurl, stream);
               curl_free(nurl);
             case CURLUE_NO_SCHEME:
             case CURLUE_NO_USER:
@@ -552,8 +647,7 @@ static void get(struct option *op, CURLU *uh)
               /* silently ignore */
               break;
             default:
-              fprintf(stderr, PROGNAME ": %s (%s)\n", curl_url_strerror(rc),
-                      v->name);
+              warnf("%s (%s)\n", curl_url_strerror(rc), v->name);
               break;
             }
           }
@@ -627,8 +721,8 @@ static const struct var *setone(CURLU *uh, const char *setline)
   return v;
 }
 
-static void set(CURLU *uh,
-                struct option *o)
+static unsigned int set(CURLU *uh,
+                        struct option *o)
 {
   struct curl_slist *node;
   unsigned int mask = 0;
@@ -642,6 +736,7 @@ static void set(CURLU *uh,
       mask |= (1 << v->part);
     }
   }
+  return mask; /* the set components */
 }
 
 static void jsonString(FILE *stream, const char *in, size_t len,
@@ -675,7 +770,7 @@ static void jsonString(FILE *stream, const char *in, size_t len,
       fputs("\\t", stream);
       break;
     default:
-      if (*i < 32)
+      if(*i < 32)
         fprintf(stream, "u%04x", *i);
       else {
         char out = *i;
@@ -694,39 +789,46 @@ static void json(struct option *o, CURLU *uh)
 {
   int i;
   bool first = true;
-  (void)o;
-  printf("%s  {\n", o->urls?",\n":"");
+  char *url;
+  CURLUcode rc = geturlpart(o, 0, uh, CURLUPART_URL, &url);
+  if(rc) {
+    VERIFY(o, ERROR_BADURL, "invalid url [%s]", curl_url_strerror(rc));
+    return;
+  }
+  printf("%s\n  {\n    \"url\": ", o->urls ? "," : "");
+  jsonString(stdout, url, strlen(url), false);
+  curl_free(url);
+  fputs(",\n    \"parts\": {\n", stdout);
   for(i = 0; variables[i].name; i++) {
-    char *nurl;
-    CURLUcode rc = curl_url_get(uh, variables[i].part, &nurl,
-                                (i?CURLU_DEFAULT_PORT:0)|
-                                CURLU_URLDECODE);
+    char *part;
+    rc = geturlpart(o, 0, uh, variables[i].part, &part);
     if(!rc) {
       if(!first)
         fputs(",\n", stdout);
       first = false;
-      printf("    \"%s\": ", variables[i].name);
-      jsonString(stdout, nurl, strlen(nurl), false);
-      curl_free(nurl);
+      printf("      \"%s\": ", variables[i].name);
+      jsonString(stdout, part, strlen(part), false);
+      curl_free(part);
     }
   }
+  fputs("\n    }", stdout);
   first = true;
   if(nqpairs) {
-    int i;
+    int j;
     fputs(",\n    \"params\": [\n", stdout);
-    for(i = 0 ; i < nqpairs; i++) {
-      const char *sep = strchr(qpairsdec[i], '=');
+    for(j = 0 ; j < nqpairs; j++) {
+      const char *sep = strchr(qpairsdec[j], '=');
       const char *value = sep ? sep + 1 : "";
 
       /* don't print out empty/trimmed values */
-      if(!qpairsdec[i][0])
-          continue;
+      if(!qpairsdec[j][0])
+        continue;
       if(!first)
         fputs(",\n", stdout);
       first = false;
       fputs("      {\n        \"key\": ", stdout);
-      jsonString(stdout, qpairsdec[i],
-                 sep ? (size_t)(sep - qpairsdec[i]) : strlen(qpairsdec[i]),
+      jsonString(stdout, qpairsdec[j],
+                 sep ? (size_t)(sep - qpairsdec[j]) : strlen(qpairsdec[j]),
                  false);
       fputs(",\n        \"value\": ", stdout);
       jsonString(stdout, value, strlen(value), false);
@@ -925,14 +1027,15 @@ static void sortquery(struct option *o)
   }
 }
 
-static CURLUcode urlfromstring(struct option *o,
-                               CURLU *uh,
-                               const char *url)
+static CURLUcode seturl(struct option *o, CURLU *uh, const char *url)
 {
   return curl_url_set(uh, CURLUPART_URL, url,
-                      CURLU_GUESS_SCHEME|CURLU_NON_SUPPORT_SCHEME|
+                      (o->no_guess_scheme ?
+                       0 : CURLU_GUESS_SCHEME)|
+                      CURLU_NON_SUPPORT_SCHEME|
                       (o->accept_space ?
-                       (CURLU_ALLOW_SPACE|CURLU_URLENCODE) : 0));
+                       CURLU_ALLOW_SPACE : 0)|
+                      CURLU_URLENCODE);
 }
 
 static void singleurl(struct option *o,
@@ -946,15 +1049,20 @@ static void singleurl(struct option *o,
     if(!uh)
       errorf(ERROR_MEM, "out of memory");
     if(url) {
-      CURLUcode rc = urlfromstring(o, uh, url);
+      CURLUcode rc = seturl(o, uh, url);
       if(rc) {
+        curl_url_cleanup(uh);
         VERIFY(o, ERROR_BADURL, "%s [%s]", curl_url_strerror(rc), url);
         return;
       }
-      else {
-        if(o->redirect)
-          curl_url_set(uh, CURLUPART_URL, o->redirect,
-                       CURLU_GUESS_SCHEME|CURLU_NON_SUPPORT_SCHEME);
+      if(o->redirect) {
+        rc = seturl(o, uh, o->redirect);
+        if(rc) {
+          curl_url_cleanup(uh);
+          VERIFY(o, ERROR_BADURL, "invalid redirection: %s [%s]",
+                 curl_url_strerror(rc), o->redirect);
+          return;
+        }
       }
     }
   }
@@ -962,8 +1070,10 @@ static void singleurl(struct option *o,
     char iterbuf[1024];
     struct curl_slist *p;
     bool url_is_invalid = false;
+    unsigned setmask = 0;
+
     /* set everything */
-    set(uh, o);
+    setmask = set(uh, o);
 
     if(iter) {
       /* "part=item1 item2 item2" */
@@ -994,6 +1104,9 @@ static void singleurl(struct option *o,
           errorf(ERROR_ITER, "bad component for iterate");
         if(iinfo->varmask & (1<<v->part))
           errorf(ERROR_ITER, "duplicate component for iterate: %s", v->name);
+        if(setmask & (1 << v->part))
+          errorf(ERROR_ITER, "duplicate --iterate and --set for component %s",
+                 v->name);
       }
       else {
         part = iinfo->part;
@@ -1074,7 +1187,7 @@ static void singleurl(struct option *o,
         url_is_invalid = true;
       }
       else {
-        rc = urlfromstring(o, uh, ourl);
+        rc = seturl(o, uh, ourl);
         if(rc) {
           VERIFY(o, ERROR_BADURL, "%s [%s]", curl_url_strerror(rc),
                  ourl);
@@ -1083,12 +1196,12 @@ static void singleurl(struct option *o,
         else {
           char *nurl = NULL;
           rc = curl_url_get(uh, CURLUPART_URL, &nurl, 0);
-          if(rc || strcmp(ourl, nurl)) {
+          if(!rc)
+            curl_free(nurl);
+          else {
             VERIFY(o, ERROR_BADURL, "url became invalid");
             url_is_invalid = true;
           }
-          if(!rc)
-            curl_free(nurl);
         }
         curl_free(ourl);
       }
@@ -1107,7 +1220,7 @@ static void singleurl(struct option *o,
     else {
       /* default output is full URL */
       char *nurl = NULL;
-      if(!curl_url_get(uh, CURLUPART_URL, &nurl, CURLU_NO_DEFAULT_PORT)) {
+      if(!geturlpart(o, 0, uh, CURLUPART_URL, &nurl)) {
         printf("%s\n", nurl);
         curl_free(nurl);
       }
@@ -1154,25 +1267,41 @@ int main(int argc, const char **argv)
     o.qsep = "&";
 
   if(o.jsonout)
-    fputs("[\n", stdout);
+    putchar('[');
 
   if(o.url) {
     /* this is a file to read URLs from */
     char buffer[4096]; /* arbitrary max */
-    while(fgets(buffer, sizeof(buffer), o.url)) {
+    bool end_of_file = false;
+    while(!end_of_file && fgets(buffer, sizeof(buffer), o.url)) {
       char *eol = strchr(buffer, '\n');
       if(eol && (eol > buffer)) {
         if(eol[-1] == '\r')
           /* CRLF detected */
           eol--;
       }
-      else if(feof(o.url))
+      else if(eol == buffer) {
+        /* empty line */
+        continue;
+      }
+      else if(feof(o.url)) {
         /* end of file */
         eol = strlen(buffer) + buffer;
+        end_of_file = true;
+      }
       else {
-        /* no newline but not end of file means that this line is truncated
-           and we are lost */
-        break;
+        /* line too long */
+        int ch;
+        warnf("skipping long line");
+        do {
+          ch = getc(o.url);
+        } while(ch != EOF && ch != '\n');
+        if(ch == EOF) {
+          if(ferror(o.url))
+            warnf("getc: %s", strerror(errno));
+          end_of_file = true;
+        }
+        continue;
       }
 
       /* trim trailing spaces and tabs */
@@ -1188,6 +1317,9 @@ int main(int argc, const char **argv)
         singleurl(&o, buffer, &iinfo, o.iter_list);
       }
     }
+
+    if(!end_of_file && ferror(o.url))
+      warnf("fgets: %s", strerror(errno));
     if(o.urlopen)
       fclose(o.url);
   }
@@ -1211,7 +1343,7 @@ int main(int argc, const char **argv)
     } while(node);
   }
   if(o.jsonout)
-    fputs("\n]\n", stdout);
+    printf("%s]\n", o.urls ? "\n" : "");
   /* we're done with libcurl, so clean it up */
   curl_slist_free_all(o.url_list);
   curl_slist_free_all(o.set_list);
